# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import logging
import os
from subprocess import Popen, PIPE

from himl.main import ConfigRunner

from kompos.cli.parser import SubParserConfig
from kompos.hierarchical.composition_helper import PreConfigGenerator, discover_compositions, sorted_compositions
from kompos.hierarchical.terraform_config_generator import TerraformConfigGenerator
from kompos.komposconfig import (
    TERRAFORM_CONFIG_FILENAME,
    get_value_or,
    local_config_dir
)
from kompos.nix import nix_install, writeable_nix_out_path, is_nix_enabled

logger = logging.getLogger(__name__)

# Terraform subcommands that will need re-initialization
SUBCMDS_WITH_INIT = [
    'plan',
    'apply',
    'destroy',
    'import',
    'state'
]

# Terraform subcommands that use the variables file.
SUBCMDS_WITH_VARS = [
    'plan',
    'apply',
    'destroy',
    'import'
]


class TerraformParserConfig(SubParserConfig):
    def get_name(self):
        return 'terraform'

    def get_help(self):
        return 'Wrap common terraform tasks with full templated configuration support'

    def configure(self, parser):
        parser.add_argument('subcommand', help='One of the terraform commands', type=str)
        parser.add_argument(
            '--var',
            help='the output var to show',
            type=str,
            default='')
        parser.add_argument('--module',
                            help='for use with "taint", "untaint" and "import". '
                                 'The module to use. e.g.: vpc', type=str)
        parser.add_argument('--resource',
                            help='for use with "taint", "untaint" and "import".'
                                 'The resource to target. e.g.: aws_instance.nat',
                            type=str)
        parser.add_argument('--name',
                            help='for use with "import". The name or ID of the imported resource. '
                                 'e.g.: i-abcd1234',
                            type=str)
        parser.add_argument('--plan', help='for use with "show", '
                                           'show the plan instead of the statefile',
                            action='store_true')
        parser.add_argument('--state-location', help='control how the remote states are used',
                            choices=['local', 'remote', 'any'], default='any', type=str)
        parser.add_argument('--force-copy',
                            help='for use with "plan" to do force state change '
                                 'automatically during init phase',
                            action='store_true')
        parser.add_argument('--template-location',
                            help='for use with "template". The folder where to save the tf files, '
                                 'without showing',
                            type=str)
        parser.add_argument('--skip-refresh', help='for use with "plan". Skip refresh of statefile',
                            action='store_false', dest='do_refresh')
        parser.set_defaults(do_refresh=True)
        parser.add_argument('--raw-output',
                            help='for use with "plan". Show raw plan output without piping through '
                                 'terraform landscape - https://github.com/coinbase/terraform-landscape '
                                 '(if terraform landscape is not enabled in komposconfig.yaml '
                                 'this will have no impact)', action='store_true',
                            dest='raw_plan_output')
        parser.set_defaults(raw_plan_output=False)
        parser.add_argument('--path-name',
                            help='in case multiple terraform paths are defined, '
                                 'this allows to specify which one to use when running terraform',
                            type=str)
        parser.add_argument(
            '--terraform-path',
            type=str,
            default=None,
            help='Path to terraform files')
        parser.add_argument('--skip-plan',
                            help='for use with "apply"; runs terraform apply without running a plan first',
                            action='store_true')
        parser.add_argument('--auto-approve',
                            help='for use with "apply". Proceeds with the apply without'
                                 'waiting for user confirmation.',
                            action='store_true')
        parser.add_argument('--himl',
                            action='store',
                            dest='himl_args',
                            default=None,
                            help='for passing arguments to himl'
                                 '--himl="--arg1 --arg2" any himl argument is supported wrapped in quotes')
        parser.add_argument(
            'terraform_args',
            type=str,
            nargs='*',
            help='Extra terraform args')

        return parser

    def get_epilog(self):
        return '''
    Examples:
        # Create/update a new cluster with Terraform
        kompos clusters/qe1.yaml terraform plan
        kompos clusters/qe1.yaml terraform apply

        # Run Terraform apply without running a plan first
        kompos clusters/qe1.yaml terraform apply --skip-plan

        # Get rid of a cluster and all of its components
        kompos clusters/qe1.yaml terraform destroy

        # Retrieve all output from a previously created Terraform cluster
        kompos clusters/qe1.yaml terraform output

        # Retrieve a specific output from a previously created Terraform cluster
        kompos clusters/qe1.yaml terraform output --var nat_public_ip

        # Refresh a statefile (no longer part of plan)
        kompos clusters/qe1.yaml terraform refresh

        # Taint a resource- forces a destroy, then recreate on next plan/apply
        kompos clusters/qe1.yaml terraform taint --module vpc --resource aws_instance.nat

        # Untaint a resource
        kompos clusters/qe1.yaml terraform untaint --module vpc --resource aws_instance.nat

        # Show the statefile in human-readable form
        kompos clusters/qe1.yaml terraform show

        # Show the plan in human-readable form
        kompos clusters/qe1.yaml terraform show --plan

        # View parsed jinja on the terminal
        kompos clusters/qe1.yaml terraform template

        # Import an unmanaged existing resource to a statefile
        kompos clusters/qe1.yaml terraform import --module vpc --resource aws_instance.nat --name i-abcd1234

        # Use the Terraform Console on a cluster
        kompos clusters/qe1.yaml terraform console

        # Validate the syntax of Terraform files
        kompos clusters/qe1.yaml terraform validate

        # Specify which terraform path to use
        kompos clusters/qe1.yaml terraform plan --path-name terraformFolder1

        # Run terraform v2 integration
        kompos data/env=dev/region=va6/project=ee/cluster=experiments terraform plan
        '''


class TerraformRunner:
    def __init__(self, root_dir, cluster_config_path, kompos_config, execute):
        self.cluster_config_path = cluster_config_path
        self.root_dir = root_dir
        self.kompos_config = kompos_config
        self.execute = execute

    def run(self, args, extra_args):
        if not os.path.isdir(self.cluster_config_path):
            raise Exception("Provide a valid composition directory path.")

        # Stop processing if an incompatible version is detected.
        validate_terraform_version(self.kompos_config.terraform_version())
        logger.info("Found extra_args %s", extra_args)

        reverse = ("destroy" == args.subcommand)

        composition_order = self.kompos_config.terraform_composition_order()
        compositions = self.check_compositions(composition_order, reverse)

        return self.run_compositions(args, extra_args, self.cluster_config_path, compositions)

    def check_compositions(self, composition_order, reverse, composition_type="composition"):
        logging.basicConfig(level=logging.INFO)

        compositions = discover_compositions(self.cluster_config_path, composition_type=composition_type)
        compositions = sorted_compositions(compositions, composition_order, reverse)

        if not compositions:
            raise Exception(
                "No terraform compositions were detected in {}.".format(self.cluster_config_path))

        return compositions

    def get_composition_path(self, nix, cloud_type, raw_config, custom):
        # Use the default local repo (not versioned).
        path = os.path.join(
            self.kompos_config.terraform_local_path(),
            self.kompos_config.terraform_root_path(),
            cloud_type,
        )

        # Overwrite with the nix output, if the nix integration is enabled.
        if nix:
            pname = self.kompos_config.terraform_repo_name()

            nix_install(
                pname,
                self.kompos_config.terraform_repo_url(),
                get_value_or(raw_config, 'infrastructure/terraform/version', 'master'),
                get_value_or(raw_config, 'infrastructure/terraform/sha256'),
            )

            # Nix store is read-only, and terraform doesn't work properly outside
            # of the module directory, so as a workaround we're using a temporary directory
            # with the contents of the derivation so terraform can create new files.
            # See: https://github.com/hashicorp/terraform/issues/18030
            path = os.path.join(
                writeable_nix_out_path(pname),
                self.kompos_config.terraform_root_path(),
                cloud_type,
            )

        if custom:
            path = os.path.join(
                path, "custom")

        return path

    def run_compositions(self, args, extra_args, config_path, compositions, custom=False):
        return_code = 0

        nix = is_nix_enabled(args, self.kompos_config.nix())

        for composition in compositions:
            logger.info("Running composition: %s", composition)

            if composition == "custom" and args.skip_custom_composition:
                logger.info("Custom compositions skipped")
                break
            elif composition == "custom" and not args.skip_custom_composition:
                logger.info("Run custom compositions")
                if "/composition=custom" not in self.cluster_config_path:
                    custom_path = self.cluster_config_path + "/composition=custom"
                else:
                    custom_path = self.cluster_config_path
                custom_compositions = discover_compositions(custom_path, composition_type="type")
                self.run_compositions(args, extra_args, custom_path, custom_compositions, custom=True)

                break

            filtered_output_keys = self.kompos_config.filtered_output_keys(composition)
            excluded_config_keys = self.kompos_config.excluded_config_keys(composition)
            pre_config_generator = PreConfigGenerator(excluded_config_keys, filtered_output_keys)
            raw_config = pre_config_generator.pre_generate_config(config_path, composition)
            cloud_type = raw_config["cloud"]["type"]

            # Generate output paths for configs
            terraform_composition_path = self.get_composition_path(nix, cloud_type, raw_config, custom=custom)

            parser = ConfigRunner.get_parser(argparse.ArgumentParser())
            if args.himl_args:
                himl_args = parser.parse_args(args.himl_args.split())
                logger.info("Extra himl arguments: %s", args.himl_args.split())
            else:
                himl_args = parser.parse_args([])

            tf_config_generator = TerraformConfigGenerator(excluded_config_keys, filtered_output_keys)

            # Generate configs
            tf_config_generator.generate_files(
                himl_args,
                config_path,
                terraform_composition_path,
                composition,
                raw_config
            )

            # Run terraform
            return_code = self.execute(
                self.run_terraform(
                    args,
                    extra_args,
                    terraform_composition_path,
                    composition
                )
            )

            if return_code != 0:
                logger.error(
                    "Command finished with nonzero exit code for composition '%s'."
                    "Will skip remaining compositions.", composition
                )
                return return_code

        return return_code

    def run_terraform(self, args, extra_args, terraform_path, composition):
        terraform_composition_path = os.path.join(terraform_path, composition)

        var_file = '-var-file="{}"'.format(TERRAFORM_CONFIG_FILENAME) if args.subcommand in SUBCMDS_WITH_VARS else ''
        terraform_env_config = 'export TF_PLUGIN_CACHE_DIR="{}"'.format(local_config_dir())

        cmd = "cd {terraform_path} && " \
              "{remove_local_cache} " \
              "{env_config} ; terraform init && terraform {subcommand} {var_file} {tf_args} {extra_args}".format(
                terraform_path=terraform_composition_path,
                remove_local_cache=remove_local_cache_cmd(args.subcommand),
                subcommand=args.subcommand,
                extra_args=' '.join(extra_args),
                tf_args=' '.join(args.terraform_args),
                var_file=var_file,
                env_config=terraform_env_config
              )

        return dict(command=cmd, post_actions=[])


def remove_local_cache_cmd(subcommand):
    if subcommand in SUBCMDS_WITH_INIT:
        return 'rm -rf .terraform &&'

    return ''


def validate_terraform_version(expected_version):
    """
    Check if the terraform binary version is compatible with the
    version specified by the kompos configuration.
    """
    try:
        execution = Popen(['terraform', '--version'],
                          stdin=PIPE,
                          stdout=PIPE,
                          stderr=PIPE)
    except Exception as e:
        logging.exception("Terraform does not appear to be installed, "
                          "please ensure terraform is in your PATH")
        exit(1)

    current_version, execution_error = execution.communicate()
    current_version = current_version.decode('utf-8').replace(
        'Terraform ', '').split('\n', 1)[0]

    if expected_version == 'latest':
        return current_version

    if current_version != expected_version and execution.returncode == 0:
        raise Exception("Terraform should be %s, but you have %s. Please change your version."\
                        % (expected_version, current_version)
        )

    return current_version



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os

from himl.main import ConfigRunner

from kompos.cli.parser import SubParserConfig
from kompos.hierarchical.composition_helper import discover_compositions
from kompos.hierarchical.config_generator import HierarchicalConfigGenerator

logger = logging.getLogger(__name__)


class ConfigGeneratorParserConfig(SubParserConfig):
    def get_name(self):
        return 'config'

    def get_help(self):
        return 'Generate configurations based on a hierarchical structure, with templating support'

    def configure(self, parser):
        return ConfigRunner().get_parser(parser)

    def get_epilog(self):
        return '''
        Examples:
        # Generate config
        kompos data/account=ee-dev/env=dev/region=va6/project=ee/cluster=experiments/composition=helmfiles config --format json --print-data
        '''


class ConfigGeneratorRunner(HierarchicalConfigGenerator):
    def __init__(self, kompos_config, cluster_config_path):
        super(ConfigGeneratorRunner, self).__init__()
        self.kompos_config = kompos_config
        self.cluster_config_path = cluster_config_path

    def run(self, args, extra_args):
        if not os.path.isdir(self.cluster_config_path):
            raise Exception("Provide a valid composition directory path.")

        config_path = os.path.join(self.cluster_config_path, "")

        compositions = discover_compositions(config_path)

        if not 0 < len(compositions) < 2:
            raise Exception("Provide the path to a single valid composition directory")
        composition = compositions[0]

        filtered = self.kompos_config.filtered_output_keys(composition).copy()
        if args.filter:
            filtered = filtered + args.filter

        excluded = self.kompos_config.excluded_config_keys(composition).copy()
        if args.exclude:
            excluded = excluded + args.exclude

        self.generate_config(
            config_path=self.cluster_config_path,
            filters=filtered,
            exclude_keys=excluded,
            enclosing_key=args.enclosing_key,
            remove_enclosing_key=args.remove_enclosing_key,
            output_format=args.output_format,
            output_file=args.output_file,
            print_data=True,
            skip_interpolation_resolving=args.skip_interpolation_resolving,
            skip_interpolation_validation=args.skip_interpolation_validation,
            skip_secrets=args.skip_secrets,
            multi_line_string=True
        )



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os

from kompos.hierarchical.config_generator import HierarchicalConfigGenerator

logger = logging.getLogger(__name__)


class PreConfigGenerator(HierarchicalConfigGenerator):

    def __init__(self, excluded_config_keys, filtered_output_keys):
        super(PreConfigGenerator, self).__init__()
        self.excluded_config_keys = excluded_config_keys
        self.filtered_output_keys = filtered_output_keys

    def pre_generate_config(self, config_path, composition, skip_secrets=True):
        return self.generate_config(
            config_path=get_config_path(config_path, composition),
            exclude_keys=self.excluded_config_keys,
            filters=self.filtered_output_keys,
            skip_interpolation_validation=True,
            skip_secrets=skip_secrets
        )


def discover_compositions(path, composition_type="composition"):
    # check single composition selected
    path_params = dict(split_path(x) for x in path.split('/'))
    composition = path_params.get(composition_type, None)
    if composition:
        return [composition]

    # discover compositions
    compositions = []
    subpaths = os.listdir(path)
    for subpath in subpaths:
        if composition_type + "=" in subpath:
            composition = split_path(subpath)[1]
            compositions.append(composition)

    return compositions


def sorted_compositions(compositions, composition_order, reverse=False):
    result = list(filter(lambda x: x in compositions, composition_order))
    return tuple(reversed(result)) if reverse else result


def split_path(value, separator='='):
    if separator in value:
        return value.split(separator)
    return [value, ""]


# Get hiera config path - source config leaf
def get_config_path(path_prefix, composition):
    prefix = os.path.join(path_prefix, '')
    if "composition=" in path_prefix:
        if "composition=custom" not in path_prefix:
            return path_prefix
        if "composition=custom" and "type=" in path_prefix:
            return path_prefix
        else:
            return "{}type={}".format(prefix, composition)
    else:
        return "{}composition={}".format(prefix, composition)


# Get target composition path - generated config
def get_composition_path(path_prefix, composition, raw_config):
    prefix = os.path.join(path_prefix, '')
    if "custom" in path_prefix:
        try:
            custom_composition = raw_config["custom"]["type"]
            logger.info("Appending custom composition: %s", custom_composition)
            return "{}{}/".format(prefix, custom_composition)
        except KeyError:
            logger.info("No custom composition type found")
            raise
    elif composition in path_prefix:
        return path_prefix
    else:
        return "{}{}/".format(prefix, composition)






# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os
from distutils.version import StrictVersion
from functools import reduce
from pathlib import Path

import fastjsonschema
import yaml

from kompos import __version__

logger = logging.getLogger(__name__)

CONFIG_SCHEMA_PATH = "data/config_schema.json"

# The filename of the generated hierarchical configuration for Terrraform.
TERRAFORM_CONFIG_FILENAME = "variables.tfvars.json"

# The filename of the generated Terrraform provider.
TERRAFORM_PROVIDER_FILENAME = "provider.tf.json"

# The filename of the generated hierarchical configuration for Helmfile.
HELMFILE_CONFIG_FILENAME = "hiera-generated.yaml"

# Directory to store terraform plugin cache
TERRAFORM_CACHE_DIR = "~/.kompos/.terraform.d/plugin-cache"


def local_config_dir(directory=TERRAFORM_CACHE_DIR):
    try:
        Path(Path.expanduser(Path(directory))).mkdir(parents=True, exist_ok=True)
        return Path.expanduser(Path(directory))

    except IOError:
        logging.error("Failed to create dir in path: %s", directory)


def get_value_or(dictionary, x_path, default=None):
    """
    Try to retrieve a value from a dictionary. Return the default if no such value is found.
    """
    keys = x_path.split("/")
    return reduce(
        lambda d, key: d.get(key, default)
        if isinstance(d, dict) else default, keys, dictionary)


class KomposConfig():
    """
    Parses all the available configuration files in order and merges them together.
    """

    DEFAULT_PATHS = [
        '/etc/kompos/.komposconfig.yaml',
        os.path.expanduser('~/.komposconfig.yaml'),
        os.path.join(os.getcwd(), '.komposconfig.yaml')
    ]

    def __init__(self, console_args, package_dir):
        cluster_config_path = console_args.cluster_config_path
        self.config = dict()
        self.package_dir = package_dir
        self.validate = fastjsonschema.compile(self.read_schema())

        paths = self.DEFAULT_PATHS[:]

        parsed_files = []
        logger.debug("parsing %s", paths)

        for config_path in paths:
            config_path = os.path.realpath(os.path.expanduser(config_path))
            if os.path.isfile(config_path):
                logger.info("parsing %s", config_path)
                with open(config_path) as f:
                    try:
                        config = yaml.safe_load(f.read())
                    except Exception as e:
                        logger.error("Failed to parse configuration file: %s", config_path)
                        raise e

                    if isinstance(config, dict):
                        parsed_files.append(config_path)
                        self.config.update(config)
                    else:
                        logger.error(
                            "cannot parse yaml dict from file: %s", config_path)

                try:
                    self.validate(config)
                except fastjsonschema.exceptions.JsonSchemaException as e:
                    logger.error("Schema validation failed for configuration file: %s", config_path)
                    raise e

        self.parsed_files = parsed_files
        logger.info("final kompos config: %s from %s", self.config, parsed_files)

    def get(self, item, default=None):
        return self.config.get(item, default)

    def validate_version(self):
        min_kompos_version = get_value_or(self.config, "min_version")

        if not min_kompos_version:
            return

        if StrictVersion(__version__) < StrictVersion(min_kompos_version):
            raise Exception(
                "The current kompos version '{}' is lower than the minimum required version '{}'".format(
                    __version__, min_kompos_version,
                )
            )

    def read_schema(self):
        with open(os.path.join(self.package_dir, CONFIG_SCHEMA_PATH), "r") as f:
            return yaml.safe_load(f.read())

    def __contains__(self, item):
        return item in self.config

    def __getitem__(self, item):
        if item not in self.config:
            raise KeyError("%s not found in %s" % (item, self.parsed_files))

        return self.config[item]

    def all(self):
        return self.config

    def nix(self):
        return get_value_or(self.config, "nix")

    def vault_backend(self):
        if get_value_or(self.config, "vault/enabled"):
            os.environ["VAULT_ADDR"] = get_value_or(self.config, "vault/url")
            os.environ["VAULT_NAMESPACE"] = get_value_or(self.config, "vault/vault_namespace")
            os.environ["VAULT_USERNAME"] = get_value_or(self.config, "vault/svc_ldap_user")
            os.environ["VAULT_ROLE"] = get_value_or(self.config, "vault/svc_ldap_user_role")
            logger.info("Vault backend enabled")

    def excluded_config_keys(self, composition, default=[]):
        return get_value_or(self.config, "compositions/config_keys/excluded/{}".format(composition), default)


    def filtered_output_keys(self, composition, default=[]):
        return get_value_or(self.config, "compositions/config_keys/filtered/{}".format(composition), default)


    def terraform_composition_order(self, default=[]):
        return self.composition_order("terraform")


    def helmfile_composition_order(self, default=[]):
        return self.composition_order("helmfile")


    def composition_order(self, composition, default=[]):
        return get_value_or(self.config, "compositions/order/{}".format(composition), default)


    def terraform_version(self):
        return get_value_or(self.config, "terraform/version", 'latest')


    def terraform_repo_url(self):
        return self.config['terraform']['repo']['url']


    def terraform_repo_name(self):
        return self.config['terraform']['repo']['name']


    def terraform_root_path(self):
        return self.config['terraform']['root_path']


    def terraform_local_path(self):
        return os.path.expanduser(self.config['terraform']['local_path'])


    def helmfile_repo_url(self):
        return self.config['helmfile']['repo']['url']


    def helmfile_repo_name(self):
        return self.config['helmfile']['repo']['name']


    def helmfile_root_path(self):
        return self.config['helmfile']['root_path']


    def helmfile_local_path(self):
        return os.path.expanduser(self.config['helmfile']['local_path'])



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import json
import logging
import os
import subprocess
import tempfile
import uuid
from distutils.dir_util import copy_tree, remove_tree
from string import Template

NIX_GIT_REPO_TEMPLATE = Template(
    """
with import <nixpkgs>{};

pkgs.stdenv.mkDerivation {
  name    = "$name";
  version = "$version";

  src = pkgs.fetchgit {
    url    = "$url";
    rev    = "$rev";
    sha256 = "$sha256";
  };

  installPhase = ''
    mkdir -p $out/src/
    cp -r * $out/src/
  '';
}
"""
)

logger = logging.getLogger(__name__)


def run_cmd(cmd):
    ret = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True,
    )

    if ret.returncode != 0:
        print(ret.stdout)
        print(ret.stderr)
        ret.check_returncode()

    return ret


def nix_generate_git_expr(name, version, info):
    """
    Generate a nix expression to install a git repository.
    """
    return NIX_GIT_REPO_TEMPLATE.safe_substitute(
        dict(
            name=name,
            version=version,
            url=info["url"],
            rev=info["rev"],
            sha256=info["sha256"],
        )
    )


def nix_install(name, repo_url, version, sha256=None):
    """
    Install a git repo with the default nix expression.
    """
    expr = nix_generate_git_expr(
        name, version, dict(url=repo_url, rev=version, sha256=sha256)
    )

    # Use nix-prefetch-git if no valid values were given.
    if not sha256 or version == "master":
        expr = nix_generate_git_expr(name, version, git_drv_info(repo_url, version))

    logging.info("Installing nix derivation %s-%s", name, version)

    with tempfile.TemporaryDirectory() as tmpdir:
        nix_file = os.path.join(tmpdir, "main.nix")

        with open(nix_file, "wb") as tmp:
            tmp.write(expr.encode("utf-8"))

        run_cmd(["nix-env", "-f", nix_file, "-i"])


def git_drv_info(url, version):
    """
    Retrieve the necessary git info to create a nix expression using `fetchgit`.
    """
    rev = []

    if version != "master":
        rev = ["--rev", version]

    ret = run_cmd(
        ["nix-prefetch-git", "--no-deepClone", "--quiet", "--url", url,] + rev
    )

    return json.loads(ret.stdout)


def nix_out_path(name):
    """
    Retrieve the path where a derivation is installed.
    """
    ret = run_cmd(["nix-env", "--query", name, "--out-path"])

    parts = list(filter(lambda x: x, ret.stdout.split(" ")))

    if len(parts) != 2:
        raise Exception("Failed to retrieve out path for derivation {}".format(name))

    return os.path.join(parts[1].rstrip("\n\r"), "src")


def writeable_nix_out_path(name):
    """
    Copy a nix derivation to a temporary writeable directory.
    """
    out_path = nix_out_path(name)
    tmp_dir = os.path.join(tempfile.gettempdir(), str(uuid.uuid1()))

    logging.info("Creating writeable directory '%s'", tmp_dir)

    if os.path.exists(tmp_dir):
        remove_tree(tmp_dir)

    os.mkdir(tmp_dir)

    logging.info(
        "Copying nix derivation '%s' to a writeable location '%s'", name, tmp_dir,
    )

    copy_tree(out_path, tmp_dir)

    return tmp_dir


def is_nix_enabled(cli_opts, cfg_opt):
    """
    Decide if nix should be enabled. CLI takes precedence over configuration file.
    """
    if cli_opts.nix:
        return True

    if cli_opts.no_nix:
        return False

    return cfg_opt



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import sys

from kompos import __version__


class RootParser():
    def __init__(self, sub_parsers=None):
        """
        :type sub_parsers: list[SubParserConfig]
        """

        if sub_parsers is None:
            sub_parsers = []
        self.sub_parsers = sub_parsers

    def _get_parser(self):
        parser = argparse.ArgumentParser(
            description='Run commands against a cluster definition', prog='kompos')
        parser.add_argument(
            'cluster_config_path',
            type=str,
            help='The cluster config path cluster.yaml')
        parser.add_argument('--root-dir', type=str, help='The root of the resource tree - '
                                                         'it can be an absolute path or relative to the current dir')
        parser.add_argument('--verbose', '-v', action='count',
                            help='Get more verbose output from commands')
        parser.add_argument(
            '--nix',
            action='store_true',
            help='Enable nix integration for remote resources'
        )
        parser.add_argument(
            '--no-nix',
            action='store_true',
            help='Disable nix integration for remote resources'
        )
        parser.add_argument(
            '--skip-custom-composition',
            action='store_true',
            help='Skip custom compositions'
        )
        parser.add_argument(
            '--version',
            action='version',
            version='%(prog)s v{version}'.format(version=__version__)
        )

        subparsers = parser.add_subparsers(dest='command')

        for subparser_conf in self.sub_parsers:
            subparser_instance = subparsers.add_parser(subparser_conf.get_name(),
                                                       help=subparser_conf.get_help(),
                                                       epilog=subparser_conf.get_epilog(),
                                                       formatter_class=subparser_conf.get_formatter())
            subparser_conf.configure(subparser_instance)

        return parser

    @staticmethod
    def _check_args_for_unicode(args):
        if args is None:
            args = sys.argv

        try:
            for value in args:
                value.encode('utf-8')
        except UnicodeDecodeError as e:
            print('Invalid character in argument "{0}", most likely an "en dash", replace it with normal dash -'.format(
                e.args[1]))
            raise

    def parse_args(self, args=None):
        RootParser._check_args_for_unicode(args)
        return self._get_parser().parse_args(args)

    def parse_known_args(self, args=None):
        RootParser._check_args_for_unicode(args)
        return self._get_parser().parse_known_args(args)


class SubParserConfig():
    def get_name(self):
        pass

    def configure(self, parser):
        pass

    def get_formatter(self):
        return argparse.RawDescriptionHelpFormatter

    def get_help(self):
        return ""

    def get_epilog(self):
        return ""



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os

from kompos.hierarchical.composition_helper import get_config_path, get_composition_path
from kompos.hierarchical.config_generator import HierarchicalConfigGenerator
from kompos.komposconfig import TERRAFORM_CONFIG_FILENAME, TERRAFORM_PROVIDER_FILENAME

logger = logging.getLogger(__name__)


class TerraformConfigGenerator(HierarchicalConfigGenerator):

    def __init__(self, excluded_config_keys, filtered_output_keys):
        super(TerraformConfigGenerator, self).__init__()
        self.excluded_config_keys = excluded_config_keys
        self.filtered_output_keys = filtered_output_keys

    def generate_files(self, himl_args, config_path, composition_path, composition, raw_config):
        config_path = get_config_path(config_path, composition)
        composition_path = get_composition_path(composition_path, composition, raw_config)

        self.generate_provider_config(himl_args, config_path, composition_path)
        self.generate_variables_config(himl_args, config_path, composition_path)

    def generate_provider_config(self, himl_args, config_path, composition_path):
        output_file = os.path.join(composition_path, TERRAFORM_PROVIDER_FILENAME)
        logger.info('Generating terraform config %s', output_file)

        filters = self.filtered_output_keys.copy() + ["provider", "terraform"]

        excluded = self.excluded_config_keys.copy()
        if himl_args.exclude:
            excluded = self.excluded_config_keys.copy() + himl_args.exclude

        self.generate_config(
            config_path=config_path,
            exclude_keys=excluded,
            filters=filters,
            output_format="json",
            output_file=output_file,
            print_data=False,
            skip_interpolation_resolving=himl_args.skip_interpolation_resolving,
            skip_interpolation_validation=himl_args.skip_interpolation_validation,
            skip_secrets=himl_args.skip_secrets
        )

    def generate_variables_config(self, himl_args, config_path, composition_path):
        output_file = os.path.join(composition_path, TERRAFORM_CONFIG_FILENAME)
        logger.info('Generating terraform config %s', output_file)

        excluded = self.excluded_config_keys.copy() + ["helm", "provider"]
        filtered = self.filtered_output_keys.copy()

        self.generate_config(
            config_path=config_path,
            exclude_keys=excluded,
            filters=filtered,
            enclosing_key="config",
            output_format="json",
            output_file=os.path.expanduser(output_file),
            print_data=True,
            skip_interpolation_resolving=himl_args.skip_interpolation_resolving,
            skip_interpolation_validation=himl_args.skip_interpolation_validation,
            skip_secrets=himl_args.skip_secrets
        )



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import os
import sys

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

with open('README.md') as f:
    _readme = f.read()

_mydir = os.path.abspath(os.path.dirname(sys.argv[0]))
_requires = [r for r in open(os.path.sep.join((_mydir, 'requirements.txt')), "r").read().split('\n') if len(r) > 1]
setup(
    name='kompos',
    version='0.3.2',
    description='Kompos - k8s cloud automation',
    long_description=_readme + '\n\n',
    long_description_content_type='text/markdown',
    url='https://github.com/adobe/kompos',
    python_requires='>=3.5',
    author='Adobe',
    author_email='noreply@adobe.com',
    license='Apache2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: HTML'
    ],
    package_dir={'': 'src'},
    packages=find_packages('src'),
    package_data={
        '': ['data/config_schema.json']
    },
    install_requires=_requires,
    entry_points={
        'console_scripts': [
            'kompos = kompos.main:run'
        ]
    }
)



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

from subprocess import call, Popen, PIPE

from .cli import display

__version__ = "0.3.2"


class Executor():
    """ All cli commands usually return a dict(command=...) that will be executed by this handler"""

    def __call__(self, result, pass_trough=True, cwd=None):
        try:
            return self._execute(result, pass_trough, cwd)
        except Exception as ex:
            display(str(ex), stderr=True, color='red')
            display(
                '------- TRACEBACK ----------',
                stderr=True,
                color='dark gray')
            import traceback
            traceback.print_exc()
            display(
                '------ END TRACEBACK -------',
                stderr=True,
                color='dark gray')

    def _execute(self, result, pass_trough=True, cwd=None):
        exit_code = 0

        if not result or not isinstance(result, dict):
            return

        if 'command' in result:
            shell_command = result['command']
            display(
                "%s" %
                shell_command,
                stderr=True,
                color='yellow')
            if pass_trough:
                exit_code = call(shell_command, shell=True, cwd=cwd)
            else:
                p = Popen(
                    shell_command,
                    shell=True,
                    stdout=PIPE,
                    stderr=PIPE,
                    cwd=cwd)
                output, errors = p.communicate()
                display(str(output))
                if errors:
                    display(
                        "%s" %
                        shell_command,
                        stderr=True,
                        color='red')
                exit_code = p.returncode

        if 'post_actions' in result:
            for callback in result['post_actions']:
                callback()

        return exit_code
